{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 19,
   "id": "33468ab3",
   "metadata": {},
   "outputs": [],
   "source": [
    "import os\n",
    "import json\n",
    "import numpy as np\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.optim as optim\n",
    "from torch.utils.data import Dataset, DataLoader\n",
    "from tqdm.auto import tqdm\n",
    "import cv2\n",
    "import matplotlib.pyplot as plt\n",
    "from IPython.display import clear_output\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "id": "e4d82a70",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 데이터셋을 불러오는 함수를 정의합니다.\n",
    "class CustomDataset(Dataset):\n",
    "    def __init__(self, dataset):\n",
    "        self.dataset = dataset\n",
    "\n",
    "    def __len__(self):\n",
    "        return len(self.dataset)\n",
    "\n",
    "    def __getitem__(self, idx):\n",
    "        image_path = self.dataset[idx][\"image_path\"]\n",
    "        mask_path = self.dataset[idx][\"mask_path\"]\n",
    "\n",
    "        image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)\n",
    "        mask = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)\n",
    "\n",
    "        # Convert mask to binary values (0 or 1)\n",
    "        _, mask = cv2.threshold(mask, 128, 1, cv2.THRESH_BINARY)\n",
    "\n",
    "        image = torch.tensor(image, dtype=torch.float32).unsqueeze(0)\n",
    "        mask = torch.tensor(mask, dtype=torch.long)\n",
    "\n",
    "        return image, mask\n",
    "        return image, mask"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "id": "c1adf79d",
   "metadata": {},
   "outputs": [],
   "source": [
    "# U-Net 모델을 구현합니다.\n",
    "class UNet(nn.Module):\n",
    "    def __init__(self, in_channels, out_channels):\n",
    "        super(UNet, self).__init__()\n",
    "\n",
    "        # Contracting Path (Encoder)\n",
    "        self.encoder1 = nn.Sequential(\n",
    "            nn.Conv2d(in_channels, 64, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(64, 64, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "        self.pool1 = nn.MaxPool2d(2)\n",
    "        \n",
    "        self.encoder2 = nn.Sequential(\n",
    "            nn.Conv2d(64, 128, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(128, 128, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "        self.pool2 = nn.MaxPool2d(2)\n",
    "\n",
    "        self.encoder3 = nn.Sequential(\n",
    "            nn.Conv2d(128, 256, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(256, 256, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "        self.pool3 = nn.MaxPool2d(2)\n",
    "\n",
    "        # Bottleneck\n",
    "        self.bottleneck = nn.Sequential(\n",
    "            nn.Conv2d(256, 512, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(512, 512, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "\n",
    "        # Expanding Path (Decoder)\n",
    "        self.upconv3 = nn.ConvTranspose2d(512, 256, 2, stride=2)\n",
    "        self.decoder3 = nn.Sequential(\n",
    "            nn.Conv2d(512, 256, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(256, 256, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "\n",
    "        self.upconv2 = nn.ConvTranspose2d(256, 128, 2, stride=2)\n",
    "        self.decoder2 = nn.Sequential(\n",
    "            nn.Conv2d(256, 128, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(128, 128, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "\n",
    "        self.upconv1 = nn.ConvTranspose2d(128, 64, 2, stride=2)\n",
    "        self.decoder1 = nn.Sequential(\n",
    "            nn.Conv2d(128, 64, 3, padding=1),\n",
    "            nn.ReLU(inplace=True),\n",
    "            nn.Conv2d(64, 64, 3, padding=1),\n",
    "            nn.ReLU(inplace=True)\n",
    "        )\n",
    "        self.final_conv = nn.Conv2d(64, out_channels, 1)\n",
    "\n",
    "    def forward(self, x):\n",
    "        enc1 = self.encoder1(x)\n",
    "        enc2 = self.encoder2(self.pool1(enc1))\n",
    "        enc3 = self.encoder3(self.pool2(enc2))\n",
    "\n",
    "        bottleneck = self.bottleneck(self.pool3(enc3))\n",
    "\n",
    "        dec3 = self.decoder3(torch.cat((enc3, self.upconv3(bottleneck)), 1))\n",
    "        dec2 = self.decoder2(torch.cat((enc2, self.upconv2(dec3)), 1))\n",
    "        dec1 = self.decoder1(torch.cat((enc1, self.upconv1(dec2)), 1))\n",
    "\n",
    "        final_output = self.final_conv(dec1)\n",
    "        return final_output"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "id": "09a6e2ff",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 데이터셋을 불러오기 위해 dataset.json 파일의 경로를 지정해줍니다.\n",
    "dataset_path = \"d:\\\\user\\\\dataset.json\"\n",
    "model_path = \"d:\\\\user\\\\model.pth\""
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "id": "0c685203",
   "metadata": {},
   "outputs": [],
   "source": [
    "#데이터셋을 불러옵니다.\n",
    "with open(dataset_path, \"r\") as f:\n",
    "    dataset = [json.loads(line) for line in f]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "id": "39fd2cff",
   "metadata": {},
   "outputs": [],
   "source": [
    "#데이터셋을 훈련셋, 검증셋, 테스트셋으로 분할합니다.\n",
    "num_samples = len(dataset)\n",
    "train_size = int(num_samples * 0.8)\n",
    "val_size = int(num_samples * 0.1)\n",
    "test_size = num_samples - train_size - val_size\n",
    "     \n",
    "train_data, val_data, test_data = torch.utils.data.random_split(dataset, [train_size, val_size, test_size])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "id": "67a5f797",
   "metadata": {},
   "outputs": [],
   "source": [
    "#데이터 로더를 초기화합니다.\n",
    "batch_size = 16\n",
    "train_loader = DataLoader(CustomDataset(train_data), batch_size=batch_size, shuffle=True)\n",
    "val_loader = DataLoader(CustomDataset(val_data), batch_size=batch_size, shuffle=True)\n",
    "test_loader = DataLoader(CustomDataset(test_data), batch_size=batch_size, shuffle=True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "id": "a0596469",
   "metadata": {},
   "outputs": [],
   "source": [
    "#모델을 불러옵니다.\n",
    "model = UNet(1, 2)\n",
    "model.load_state_dict(torch.load(model_path))\n",
    "\n",
    "def evaluate_model(model, data_loader, device):\n",
    "#     model.to(device) # 모델을 device로 보냅니다.\n",
    "#     loss_total = 0\n",
    "#     accuracy_total = 0\n",
    "#     model.eval()  # 모델을 평가 모드로 설정합니다.\n",
    "\n",
    "    model.eval()\n",
    "    loss_total = 0.0\n",
    "    correct = 0\n",
    "    total = 0\n",
    "\n",
    "    # 손실 함수를 device로 보냅니다.\n",
    "    loss_fn = nn.CrossEntropyLoss().to(device)\n",
    "\n",
    "    with torch.no_grad():\n",
    "        for images, labels in data_loader:\n",
    "            # GPU로 이미지와 레이블을 전송합니다.\n",
    "            images = images.to(device)\n",
    "            labels = labels.to(device)\n",
    "            \n",
    "            # 모델에 이미지를 입력합니다.\n",
    "            outputs = model(images)\n",
    "\n",
    "            # 손실 함수를 사용해 손실을 계산합니다.\n",
    "            loss = loss_fn(outputs, labels)\n",
    "\n",
    "            # 손실 값을 기록합니다.\n",
    "            loss_total += loss.item() * images.size(0)\n",
    "\n",
    "            # 정확도를 계산합니다.\n",
    "            _, predicted = torch.max(outputs.data, 1)\n",
    "            total += labels.size(0)\n",
    "            correct += (predicted == labels).sum().item()\n",
    "\n",
    "\n",
    "    # 손실과 정확도를 계산합니다.\n",
    "    loss_avg = loss_total / len(data_loader.dataset)\n",
    "    acc_avg = correct / total\n",
    "\n",
    "\n",
    "    return loss_avg, acc_avg\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "id": "a80edf03",
   "metadata": {},
   "outputs": [],
   "source": [
    "#이제 위의 train_model 함수를 사용하여 모델을 학습시킬 수 있습니다. 이 함수를 사용하려면, 이전에 정의한 모델, 데이터 로더, 손실 함수, 최적화기, 디바이스, 에포크 수 및 저장 간격을 인자로 전달해야 합니다\n",
    "#train_model(model, train_loader, criterion, optimizer, device, num_epochs, save_interval)\n",
    "#이렇게 하면, 주어진 에포크 수에 대해 전체 학습 과정이 완료됩니다. 이후 평가 과정 및 모델 저장이 필요할 수 있습니다.\n",
    "\n",
    "# def train_model(model, train_loader, criterion, optimizer, device, num_epochs, save_interval):\n",
    "#     model.train()\n",
    "    \n",
    "#     for epoch in range(num_epochs):\n",
    "#         for i, (images, labels) in enumerate(train_loader):\n",
    "#             # GPU로 이미지와 레이블을 옮기기\n",
    "#             images = images.to(device)\n",
    "#             labels = labels.to(device)\n",
    "    \n",
    "#             # Forward pass\n",
    "#             outputs = model(images)\n",
    "    \n",
    "#             # 손실 함수 계산\n",
    "#             loss = criterion(outputs, labels)\n",
    "    \n",
    "#             # Backward pass 및 최적화\n",
    "#             optimizer.zero_grad()\n",
    "#             loss.backward()\n",
    "#             optimizer.step()\n",
    "    \n",
    "#             # 일정 간격으로 학습 진행 상황을 출력\n",
    "#             if (i + 1) % save_interval == 0:\n",
    "#                 print(f\"Epoch [{epoch + 1}/{num_epochs}], Step [{i + 1}/{len(train_loader)}], Loss: {loss.item():.4f}\")\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 28,
   "id": "25614295",
   "metadata": {},
   "outputs": [],
   "source": [
    "def train_model(model, train_loader, val_loader, criterion, optimizer, checkpoint_dir, num_epochs, device, start_epoch=0):\n",
    "    loss_values = []\n",
    "    train_acc_values = []\n",
    "    val_acc_values = []\n",
    "    best_val_loss = float(\"inf\")\n",
    "    save_interval = 100\n",
    "    checkpoint_interval = 5\n",
    "\n",
    "    # 모델을 device로 보냅니다.\n",
    "    model.to(device)\n",
    "\n",
    "    # 학습 루프를 설정합니다.\n",
    "    for epoch in range(start_epoch, num_epochs):\n",
    "        # 학습 데이터셋에 대한 손실과 정확도를 계산합니다.\n",
    "        train_loss, train_accuracy = evaluate_model(model, train_loader, device)\n",
    "\n",
    "        # 검증 데이터셋에 대한 손실과 정확도를 계산합니다.\n",
    "        val_loss, val_accuracy = evaluate_model(model, val_loader, device)\n",
    "\n",
    "        # 손실 값을 출력합니다.\n",
    "        print(f\"Epoch {epoch + 1} (train) - Loss: {train_loss:.4f}, Accuracy: {train_accuracy:.4f}\")\n",
    "        print(f\"Epoch {epoch + 1} (val) - Loss: {val_loss:.4f}, Accuracy: {val_accuracy:.4f}\")\n",
    "\n",
    "        # 손실 값을 저장하고 그래프를 그립니다.\n",
    "        loss_values.append(val_loss)\n",
    "        plot_loss_graph()\n",
    "\n",
    "        # 모델을 저장합니다.\n",
    "        if (epoch + 1) % checkpoint_interval == 0:\n",
    "            torch.save(model.state_dict(), os.path.join(checkpoint_dir, f\"model_{epoch+1}.pth\"))\n",
    "            print(f\"Checkpoint saved at epoch {epoch+1}.\")\n",
    "\n",
    "        # 학습을 수행합니다.\n",
    "        model.train()\n",
    "        for i, (images, labels) in enumerate(train_loader):\n",
    "            # GPU로 이미지와 레이블을 옮기기\n",
    "            images = images.to(device)\n",
    "            labels = labels.to(device)\n",
    "\n",
    "            # Forward pass\n",
    "            outputs = model(images)\n",
    "\n",
    "            # 손실 함수 계산\n",
    "            loss = criterion(outputs, labels)\n",
    "\n",
    "            # Backward pass 및 최적화\n",
    "            optimizer.zero_grad()\n",
    "            loss.backward()\n",
    "            optimizer.step()\n",
    "\n",
    "    # 일정 간격으로 학습 진행 상황을 출력\n",
    "    if (i + 1) % save_interval == 0:\n",
    "        print(f\"Epoch [{epoch + 1}/{num_epochs}], Step [{i + 1}/{len(train_loader)}], Loss: {loss.item():.4f}\")\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c14300c7",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.16"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
